from .util import *

def is_identifier(w):
    if not w:
        return False

    if not w[0].isalpha() and w[0] != '_':
        return False

    if w.replace('_', '').isalnum():
        return True

    return False
        
def find_identifiers(line):
    seps = list(',;.=->*+/()[]%')
    seps.append(' ')
    return smart_split(line, seps, keep_splitters=True)

# we do some touchups to remove something like foo.bar.baz
# into only foo
def add_identifiers(line, scope):
    toks = find_identifiers(line)

    new = []

    new_toks = []
    toks = ['', '']
    toks.extend(list(find_identifiers(line)))
    while toks:
        t = toks.pop()
        new_toks.append(t)

        if toks:
            nt = toks[-1]
            if nt == '.':
                toks.pop()
                new_toks.pop()

            if nt == '>':
                pt = toks[-2]
                if pt == '-':
                    toks.pop()
                    toks.pop()
                    new_toks.pop()
                

    for t in new_toks:
        if is_identifier(t):
            if t not in scope:
                new.append(t)
                scope[t] = 1
    
    return new

if __name__ == "__main__":
    scope = {}
    add_identifiers("node** tree_init()", {})



from .transforms import comments, io, keywords, structure, variables
from . import analysis
from . import util

import os

def pipeline(lines, base_dir=None):
    lines = comments.skip_comments(lines)
    lines = structure.join_backslash_lines(lines)
    lines = structure.add_preceding_ignore_chars(lines)

    lines = keywords.replace_raw(lines, base_dir or os.getcwd())
    lines = keywords.replace_tabs(lines)
    lines = keywords.replace_pass(lines)
    lines = keywords.replace_for_shorthand(lines)
    lines = keywords.replace_knowns(lines)
    lines = keywords.replace_blocks(lines)
    lines = keywords.replace_defs(lines)
    lines = keywords.replace_self(lines)

    # replaces !, ?, ??, print, read, etc
    lines = io.replace_io_keywords(lines)

    # scopings is a per line scope of seen variables
    scopings = analysis.read_scopings(lines)
    lines = variables.add_auto_declarations(lines, scopings)
    lines = variables.add_destructuring(lines, scopings)

    lines = structure.add_parentheses(lines)
    lines = structure.add_trailing_semicolons(lines)

    # add curly braces (from indentation) has to be last
    lines = structure.add_curly_braces(lines)

    # TODO: decide on whether to remove from generated code or not
    lines = structure.remove_preceding_ignore_chars(lines)

    requires = analysis.guess_required_files(lines)

    lines = requires + lines
    return lines



from ..util import *

def var_access(arg):
    return dot_access(arg) or array_access(arg) or ptr_access(arg)

def no_punctuation(arg):
    narg = []
    for c in arg:
        if c.isalnum() or c == '_':
            narg.append(c)

    return ''.join(narg)

DESTRUCTURE_INDEX = 0

DECLARE_VARIABLES = True

# this function splits equals, skipping over '==' and turning '= ' into ''
def split_equals(line):
    tokens = []
    prev = []
    i = 0
    while i < len(line):
        c = line[i]
        peeked = ''
        if c == '=':
            if i < len(line)-1:
                peeked = line[i+1]

            if peeked == '=':
                prev.append(c)
                prev.append(peeked)
                i += 1
            else:
                if prev:
                    tokens.append(''.join(prev))
                prev = []

        else:
            prev.append(c)
        i += 1

    tokens.append(''.join(prev))

    return tokens


def make_declarations(line, scope):
    global DESTRUCTURE_INDEX

    di = DESTRUCTURE_INDEX
    indent = get_indent(line)

    # we need to split on single equals but not double equals
    # we also need to swallow '= ' and turn it into '='
    tokens = split_equals(line.strip())

    # TODO: what do we do if tokens > 2?
    if len(tokens) == 2:
        lhs, rhs = tokens
        rhs = rhs.strip()
        args = smart_split(lhs, ',')
        if len(args) > 1:

            if lhs.find('[') != -1:
                return line

            need_args = False
            for arg in args:
                arg = arg.strip(',').strip()
                if arg not in scope:
                    need_args = True
            rhs = rhs.strip()

            if not need_args:
                args = ','.join(args).strip()
                line = '%sstd::tie(%s) = %s' % (' ' * indent, args, rhs)
            else:
                pname = "structuredArgs_%s" % (di)
                line = '%sauto %s = %s;' % (' ' * indent, pname, rhs)
                for j, arg in enumerate(args):
                    arg = arg.strip(',').strip()
                    if arg == '_':
                        continue

                    if not arg in scope:
                        line += '\n%sauto %s = get<%s>(%s);' % (' ' * indent, arg, j, pname )
                    else:
                        line += '\n%s%s = get<%s>(%s);' % (' ' * indent, arg, j, pname)

                di += 1

        elif len(args[0].split()) == 1:
            arg = args[0].strip()
            carg = no_punctuation(args[0])
            if not var_access(arg) and not carg in scope:
                line = "%sauto %s = %s" % (' ' * indent, arg, rhs)

    DESTRUCTURE_INDEX = di
    return line

# adds tuples to return statements and std::tie to assignments
def add_destructuring(lines, scopings):
    new_lines = []
    keywords = [ "for", "while", "do" ]
    for i, line in enumerate(lines):
        if ignore_line(line, new_lines):
            continue
        # remove any trailing ':' and whitespace
        line = line.rstrip()
        sline = line.strip()

        if i in scopings:
            scope = scopings[i]
        else:
            scope = {}

        added = False
        for k in keywords:
            if sline.startswith(k):
                new_lines.append(line)
                added = True
                break

        if added:
            continue

        if line.strip().startswith('return'):
            indent = get_indent(line)
            args = smart_split(line[indent + len('return'):], ',')
            if len(args) > 1:
                args = ', '.join(args).strip()
                line = "%sreturn make_tuple(%s)" % (' ' * indent, args)

        elif line.find('=') != -1:
            if DECLARE_VARIABLES:
                line = make_declarations(line, scope)


        new_lines.append(line)

    return new_lines

def get_class(line):
    tokens = smart_split(line, ' ')
    if tokens[0] == "class":
        return tokens[1].rstrip(':')

    return ""
def add_auto_declarations(lines, scopings):
    new_lines = []
    class_start = 0

    keywords = ["if", "do ", "while", "else", "class", "struct", "typedef"]
    in_class = ""
    for i, line in enumerate(lines):
        if ignore_line(line, new_lines):
            continue
        sline = line.strip()
        scope = scopings[i]
        indent = get_indent(line)

        if is_class(line):
            in_class = get_class(line)

        skip_line = False
        for k in keywords:
            if line.find(k) != -1:
                new_lines.append(line)
                skip_line = True
                break

        if skip_line:
            continue

        next_indent = indent
        if i < len(lines) - 1:
            next_line = lines[i+1]
            next_indent = get_indent(next_line)


        # check for function level declarations,
        # they either happen at the toplevel (0)
        # or inside a class and on the first level
        # TODO: figure out where more of these can happen
        if indent == 0 or indent < next_indent:
            type = ""
            var = ""
            pr = line.find("(")
            if pr > 0:
                lp = line.find(")")
                params = smart_split(line[pr+1:lp], ",")

                before_p = line[:pr]
                after_p = line[lp+1:]
                prev = None

                new_params = []
                # if any param has spaces in it, then we think it is a function declaration
                # foo(int a) vs foo(10, 20, "abc")
                for p in params:
                    args = smart_split(p, " ")
                    if len(args) == 1:
                        var = args[0]
                    elif len(args) >= 2:
                        type = " ".join(args[:-1])
                        var = args[-1]

                    if type:
                        new_params.append("%s %s" % (type, var))
                    else:
                        new_params.append(var)

                args = before_p.split()
                class_func = False

                if before_p.strip() == in_class.strip():
                    class_func = True
                elif before_p.strip() == "~%s" % in_class.strip():
                    class_func = True

                if before_p == "main":
                    if len(args) == 1:
                        line = "%sint %s(%s)%s" % (' ' * indent, before_p, ", ".join(new_params), after_p)
                elif len(args) == 1 and not class_func:
                    before_p = before_p.strip()
                    line = "%sauto %s(%s)%s" % (' ' * indent, before_p, ", ".join(new_params), after_p)
                else:
                    line = "%s(%s)%s" % (before_p, ", ".join(new_params), after_p)


        # special for loop declarations
        if sline.startswith('for ') or sline.startswith('for('):
            # if its a for loop, remove ending colon just in case so we don't
            # swallow it when parenthesizing
            sline = sline.rstrip(':')

            args = smart_split(sline, ';')
            if sline.startswith('for('):
                args[0] = args[0][len('for'):]
            else:
                args[0] = args[0][len('for '):]

            args[0] = strip_outer_parens(args[0])
            stmts = smart_split(args[0], ',')

            arg0 = []
            for j, s in enumerate(stmts):
                toks = s.split('=')
                if len(toks) == 1:
                    arg0.append(s)
                    continue

                lhs, rhs = toks
                rhs = rhs.strip()

                arg = lhs.strip()

                add_auto = True
                if j != 0:
                    add_auto = False
                elif var_access(arg):
                    add_auto = False
                elif arg in scope:
                    add_auto = False
                elif len(arg.split()) > 1:
                    add_auto = False

                if add_auto:
                    arg = "auto %s" % (arg)

                arg0.append("%s = %s" % (arg, rhs))

            args[0] = ", ".join(arg0)
            line = "%sfor %s" % (' '*indent, ";".join(args))



        new_lines.append(line)
    return new_lines

